from uologging.uologging import init_console_logging, \
    set_logging_verbosity, init_syslog_logging
from uologging.cli import get_default_parser, add_verbosity_flag
from uologging.performance import trace


__all__ = [
    'trace',
    'init_console_logging', 
    'init_syslog_logging',
    'set_logging_verbosity', 
    'get_default_parser', 
    'add_verbosity_flag', 
]



import argparse


def add_verbosity_flag(parser: argparse.ArgumentParser):
    """Add a --verbose/-v flag to an argparse.ArgumentParser.

    Args:
        parser (argparse.ArgumentParser): The argument parser to be updated.

    Example:
        This should be invoked like the following:
        >>> import uologging
        >>> import argparse
        >>> parser = argparse.ArgumentParser()
        >>> uologging.add_verbosity_flag(parser)

        When parsing args, the verbosity_flag will be set if --verbose/-v is passed!
        >>> args = parser.parse_args(['-vv'])
        >>> args.verbosity_flag
        2

        The verbosity_flag is intended to be used as input for set_logging_verbosity().
    """
    parser.add_argument(
        '-v', '--verbose',
        dest='verbosity_flag',
        action='count',
        default=0,
        help='Logging verbosity. "-vv" results in maximum logging.'
    )


def get_default_parser():
    """Provides an argparse ArgumentParser configured to take --verbose/-v flag.

    Returns:
        argparse.ArgumentParser: Provides a --verbose/-v flag. Should be used as a 'parent parser.'

    Example:
        This should be invoked like the following:
        >>> import uologging
        >>> import argparse
        >>> parser = argparse.ArgumentParser(parents=[uologging.get_default_parser()])

        The parser has been configured with the uologging default_parser as a parent parser.
        When parsing args, the verbosity_flag will be set if --verbose/-v is passed!
        >>> args = parser.parse_args(['-vv'])
        >>> args.verbosity_flag
        2

        The verbosity_flag is intended to be used as input for set_logging_verbosity().
    """
    parser = argparse.ArgumentParser(add_help=False)
    add_verbosity_flag(parser)
    return parser



import time
import logging
from functools import wraps


def trace(logger: logging.Logger, capture_args=True):
    """Decorator to trace the execution time of a Python function or method.

    Args:
        logger (logging.Logger): The logger instance for the function's module.
        capture_args (bool, optional): When tracing the function, capture the 
            arguments provided to the function and add them to the log message.
            Defaults to True.

    Example:

    First, get the logger for your Python module.

        import logging
        logger = logging.getLogger(__name__)

    Then, use the trace_time decorator with that logger as the argument.

        import uologging
        @uologging.trace(logger)
        def my_slow_function():
            import time
            time.sleep(1)
        my_slow_function()
    """
    def _trace_time(func):
        @wraps(func)
        def timed(*args, **kw_args):
            if capture_args:
                func_str_repr = f'{func.__module__}:{func.__name__}({args!r},{kw_args!r})'
            else:
                func_str_repr = f'{func.__module__}:{func.__name__}(...)'
            logger.debug(f'Starting: {func_str_repr}')
            start = time.time()
            result = func(*args, **kw_args)
            end = time.time()
            logger.debug(f'Finished: {func_str_repr}')
            logger.info(f'{func_str_repr} exec time: {end - start:.2f} sec')
            return result
        return timed
    return _trace_time



import logging
import logging.handlers


_logging_format = logging.Formatter(
    '%(asctime)s %(levelname)-8s  %(message)s [%(pathname)s:%(lineno)d]',
    '%Y-%m-%d %H:%M:%S')


syslog_logging_initialized = False
def init_syslog_logging():
    """Setup logging to be output to syslog (via /dev/log).

    Example:
        Call this method.
        >>> import uologging
        >>> uologging.init_syslog_logging()

        Then use the Python logging package in each of your package's modules.
        >>> import logging
        >>> logger = logging.getLogger(__name__)

        We use a hardcoded str here to enable this doctest:
        >>> logger = logging.getLogger('examplepkg.just.testing')
        >>> logger.critical('Just kidding, this is a test!')    
    """
    global syslog_logging_initialized
    if syslog_logging_initialized is False:
        syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
        _add_root_log_handler_with_formatter(syslog_handler)
        syslog_logging_initialized = True


console_logging_initialized = False
def init_console_logging():
    """Setup logging to be output to the console.

    Example:
        Call this method.
        >>> import uologging
        >>> uologging.init_console_logging()

        Then use the Python logging package in each of your package's modules.
        >>> import logging
        >>> logger = logging.getLogger(__name__)

        We use a hardcoded str here to enable doctest:
        >>> logger = logging.getLogger('examplepkg.just.testing')
        >>> logger.critical('Just kidding, this is a test!')
    """
    global console_logging_initialized
    if console_logging_initialized is False:
        console_handler = logging.StreamHandler()
        _add_root_log_handler_with_formatter(console_handler)
        console_logging_initialized = True


def set_logging_verbosity(verbosity_flag: int):
    """Set the logging verbosity for your entire package.

    This leverages the 'root logger' paradigm provided by Python logging.

    Args:
        verbosity_flag (int): Higher number means more logging. Choices are [0,2]. 
            Default is 0. Default will captures WARNING, ERROR, and CRITICAL logs.
            Provide 1 to also capture INFO logs. Provide 2 to also capture DEBUG logs.
    """
    root_logger = logging.getLogger()
    if verbosity_flag == 1:
        root_logger.setLevel(logging.INFO)
    elif verbosity_flag >= 2:
        root_logger.setLevel(logging.DEBUG)
    else:  # Default to WARNING, following Python logging standards
        root_logger.setLevel(logging.WARNING)


def _add_root_log_handler_with_formatter(handler: logging.Handler):
    """Add a handler to a the root logger for package.

    Args:
        handler (logging.Handler): A logging handler.
    """
    handler.setFormatter(_logging_format)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
